//**********************************************************************************************************\
//
//	archive_wad.c:	Wad files
//
//**********************************************************************************************************/

#include "project.h"

#define MAX_WAD_FILES_IS_4096 4096
#define MAX_WAD_FILENAME_IS_15	15

// WAD file structure
// 1. Header
// 2. File bytes one after another
// 3. Directory

#define isWAD3(myWad) (myWad->entryChar == 'C')

typedef struct
{ 
	char		id[4];				// "WAD2", Name of the new WAD format
	int			numEntries;         // Number of entries
	int			directoryOffset;    // Position of WAD directory in file
} wadheader_t;

typedef struct
{ 
	int			filepos;			// Position of the entry in WAD
	int			filelen;            // Size of the entry in WAD file
	int			memsize;            // Size of the entry in memory
	char		type;               // type of entry, should be 'D' (0x44) MipTex ('C' for Half-Life
	char		unused[3];          // Never compressed, so why bother
	char		name[16];           // 1 to 16 characters, '\0'-padded
} wadentry_t;


typedef struct wad_access_s
{
	wadheader_t		header;
	wadentry_t		directory[MAX_WAD_FILES_IS_4096];

	char			fileName[MAX_PATH_IS_1024];
	FILE*			fileHandle;
	int				fileMode;							// ARCHIVE_CREATE, ARCHIVE_EDIT, ARCHIVE_READ
	fbool			isDirty;

	fbool			isWAD3;

	int				totalEntries;
	int				dataSegmentSize;

} wad_access_t;

static const char* accessCodes[] = 
{
	"wb",	// ARCHIVE_CREATE 0
	"rb",	// ARCHIVE_READ	1
	"r+b",	// ARCHIVE_EDIT 2
	"r+b",	// ARCHIVE_FASTEDIT 3
};

#define OPEN_FILE_1 1
#define WAD_CHECK_2 2
#define DIR_SIZE_3  3

const char* wad_open_error[] =
{
	"Unknown",
	"Couldn't open file",
	"Not a wad file",
	"Too many files, 4096 max",
};

#define CHECK_MODE_1 1
#define CHECK_LENGTH_2 2
#define CHECK_NUM_FILES_3 3
#define CHECK_APPEND_4 4

const char* wad_add_error[] =
{
	"Unknown",
	"Wad mode is read-only, cannot write",
	"Filename for wad directory exceeds 15 characters",
	"Too many files in wad, 4096 max",
	"Couldn't append data from file to wad",
};

//
// SUPPORTING FUNCTIONS
//

static fbool WadFile_Close_Directory_AlphaSort (wad_access_t* myWad)
{
#if 0
	return False;
#else
	fbool	didWeSort = False;
	int		name_in_mem = (int)&myWad->directory[0].name[0] - (int)&myWad->directory[0];

	if (myWad->fileMode == ARCHIVE_READ || myWad->totalEntries == 0)
		return False; // Read can't write or crazy situation with no entries

	didWeSort = Sort_Memory_Blocks_By_Name ((char*)myWad->directory, sizeof(myWad->directory[0]), myWad->totalEntries, name_in_mem);
	
//	printf ("Did we alpha sort directory? %s ...\n", didWeSort ? "Yes" : "No" );
	return didWeSort; // We might have sorted.  We might have not needed to.
#endif
}

static fbool WadFile_AlphaSort_Test (wad_access_t* myWad)
{
	fbool	didWeSort = False;
	int		name_in_mem = (int)&myWad->directory[0].name[0] - (int)&myWad->directory[0];

	if (myWad->totalEntries == 0)
		return False; // Read can't write or crazy situation with no entries

	didWeSort = Sort_Memory_Blocks_By_Name ((char*)myWad->directory, sizeof(myWad->directory[0]), myWad->totalEntries, name_in_mem);
	
	return didWeSort; // We might have sorted.  We might have not needed to.
}


// Look through the pack and see if there is empty space
static fbool WadFile_Close_Check_For_Empty_Space (wad_access_t* myWad)
{
	// Assumes cursor is at the right place.  Create, Edit, FastEdit this is always true.  Read?  Well, make it so.
	int		directoryStart		= File_GetCursor (myWad->fileHandle);	
	int		expectedDataSize	= directoryStart - sizeof(myWad->header);
	int		foundSize			= 0;
	fbool	isEmptySpace		= False;
	int		i;

	if (myWad->fileMode == ARCHIVE_READ && directoryStart != myWad->header.directoryOffset)
		printf ("Hey chief, the cursor into the file seems to be in wrong place");

	for (i = 0; i < myWad->totalEntries; i ++)
		foundSize += myWad->directory[i].filelen;

	if (foundSize != expectedDataSize)
		isEmptySpace = True;

//	printf ("Needs compressed? %s ...\n", isEmptySpace ? "Yes" : "No" );
	return isEmptySpace;
}

// Removes directory entries of deleted files
static fbool WadFile_Close_Directory_Fixup (wad_access_t* myWad)
{
#if 0
	return False;
#else

	fbool didAnything = False;
	int	newNumEntries = 0;
	int	i, j;
	
	if (myWad->fileMode == ARCHIVE_CREATE || myWad->fileMode == ARCHIVE_READ)
		return False; // "Create" doesn't support delete and read doesn't write at all

	// Determine actual number of entries
	for (i = 0; i < myWad->totalEntries; i ++)
		if (myWad->directory[i].name[0])
			newNumEntries ++; 

	// Recover space moving entries
	for (i = 0; i < newNumEntries; i ++)
		if (!myWad->directory[i].name[0]) // Is blank entry?
			for (j = i + 1;  i < myWad->totalEntries; i ++) // // Find a non-blank entry and move it in there
				if (myWad->directory[j].name[0])
				{
					// Relocate and erase
					memcpy (&myWad->directory[i], &myWad->directory[j], sizeof(myWad->directory[i]) );
					memset (&myWad->directory[j], 0,					  sizeof(myWad->directory[j]) );
					didAnything = True;
					break; // Get out
				}

	myWad->totalEntries = newNumEntries;

	printf ("Did we clean up deleted files? %s ...\n", didAnything ? "Yes" : "No" );
	return didAnything;
#endif
}


// Make a copy without the extra space (where deleted data lives)
static void WadFile_Close_Compress (wad_access_t* myWad)
{
	const char* tempFileName = "temp.tmp";
	FILE*		fout		 = fopen (tempFileName, "wb"); // Write-only
	int			cursor		 = 0;
	int			oldsize		 = myWad->header.directoryOffset - sizeof(myWad->header);
	int			i, newsize, recoveredsize;

	printf ("Compressing ...\n");
	if (!fout)
	{
		printf ("Space recover: Couldn't open temp file for writing\n");  // Let's see if this happens?
		return; // This can't be a fatal error.  Otherwise we lose additions, deletions and such.
	}

	cursor += File_Safe_Write (fout, &myWad->header, sizeof(myWad->header) );	// Just to keep things easy
	
	for (i = 0; i < myWad->totalEntries; i ++)
	{
		File_Append_FileBlock (fout, myWad->fileHandle, myWad->directory[i].filepos, myWad->directory[i].filelen);
		myWad->directory[i].filepos = cursor; // Update the file position
		cursor += myWad->directory[i].filelen;
	}

	// Write temp file directory and re-write the temp file header after refreshing it.
	{	
		int directoryStart			= File_GetCursor (fout);	
		int directoryByteLength		= sizeof(myWad->directory[0]) * myWad->totalEntries;
		newsize						= directoryStart - sizeof(myWad->header);
		recoveredsize				= oldsize - newsize;

		myWad->header.id[0]		= 'W';		// Construct final header
		myWad->header.id[1]		= 'A';
		myWad->header.id[2]		= 'D';
		myWad->header.id[3]		= myWad->isWAD3 ? '3' : '2';
		myWad->header.numEntries	  = LittleLong(myWad->totalEntries);
		myWad->header.directoryOffset = LittleLong(directoryStart);
		
		File_Safe_Write				(fout, &myWad->directory, directoryByteLength);  // Write directory
		File_SetCursor_ToStart		(fout);											// Return to beginning of file
		File_Safe_Write				(fout, &myWad->header, sizeof(myWad->header) ); // Write header
	}

	fclose (fout);					// New file completed.
	fclose (myWad->fileHandle);	// Old pack data is stale.
	
	File_Delete (myWad->fileName);
	File_Rename (tempFileName, myWad->fileName);

	myWad->fileHandle = fopen	(myWad->fileName, "rb"); // Re-open file, but in read-mode.
	printf ("Compress pack old size %i, new size %i, recovered size %i:", oldsize, newsize, recoveredsize);
}

// Browse through in-memory representation
static int WadFile_Find_File_Index (wad_access_t* myWad, const char* storedName)
{
	int i;
	
	if (storedName[0] == 0) // Sanity
		return -1;

	for (i = 0; i < myWad->totalEntries; i++)
		if (StringMatch (storedName, myWad->directory[i].name))
			return i;
	
	return -1;
}

//
// ACTUAL FUNCTIONS (EXTERNAL)
//
struct wad_access_s* WadFile_Auto_Open2 (const char* myWadFileName)
{
// If file exists, open for fast edit.  If file doesn't exist, open for create
	int myFileMode = File_Exists(myWadFileName) ? ARCHIVE_FASTEDIT : ARCHIVE_CREATE;
	return WadFile_Open (myWadFileName, myFileMode);
}

struct wad_access_s* WadFile_Auto_Open3 (const char* myWadFileName)
{
// If file exists, open for fast edit.  If file doesn't exist, open for create WAD3
	int myFileMode = File_Exists(myWadFileName) ? ARCHIVE_FASTEDIT : ARCHIVE_CREATE_WAD3;
	return WadFile_Open (myWadFileName, myFileMode);
}


struct wad_access_s* WadFile_Open (const char* myWadFileName, int myFileMode)
{
	wad_access_t*	myWad		= Memory_calloc (sizeof(wad_access_t), 1, "In-Memory Pack File Struct");
	int				realMode	= myFileMode == ARCHIVE_CREATE_WAD3 ? ARCHIVE_CREATE : myFileMode;
	const char*		accessCode = accessCodes[realMode];
	int				stage		= OPEN_FILE_1;

	if ( !(myWad->fileHandle = fopen (myWadFileName, accessCode) ) )  goto wad_open_fail; // Assign filehandle, but if NULL then error out

	StringLCopy (myWad->fileName, myWadFileName);
	myWad->fileMode = myFileMode;	// Edit.  Create, write, read handled differently
	
	if (myFileMode == ARCHIVE_CREATE_WAD3)
	{
		myFileMode = myWad->fileMode = ARCHIVE_CREATE;
		myWad->isWAD3	= True;
	}

	if (myFileMode == ARCHIVE_CREATE)  // If Create just write a dummy header and get out ...
	{
		File_Safe_Write (myWad->fileHandle, &myWad->header, sizeof(myWad->header) );
		myWad->isDirty = True;
	}
	else
	{	// ... ekse read header/directory, set cursor to start of directory (EDIT may write new data there)
		
		File_Safe_Read				(myWad->fileHandle, &myWad->header, sizeof(myWad->header) );

		stage = WAD_CHECK_2;		     if (memcmp(myWad->header.id, "WAD2", 4) == 0)					
											myWad->isWAD3 = False;
									else if (memcmp(myWad->header.id, "WAD3", 4) == 0)					
											myWad->isWAD3 = True;
									else															
											goto wad_open_fail;
		stage = DIR_SIZE_3;			if (myWad->header.numEntries > MAX_WAD_FILES_IS_4096 )			goto wad_open_fail;

		File_SetCursor				(myWad->fileHandle, myWad->header.directoryOffset);						// Set cursor
		File_Safe_Read				(myWad->fileHandle, myWad->directory, myWad->header.numEntries * sizeof(myWad->directory[0]) );	// Get directory
		File_SetCursor				(myWad->fileHandle, myWad->header.directoryOffset);						// Move cursor to beginning of directory

		myWad->totalEntries			= myWad->header.numEntries;
		myWad->dataSegmentSize		= myWad->header.directoryOffset - sizeof(myWad->header);
	}   
	return myWad;  // Everything ok.  Get out with struct calloc'd and file open ...

wad_open_fail:   

	printf ("'%s': %s", myWadFileName, wad_open_error[stage]);
	if (myWad->fileHandle)  fclose (myWad->fileHandle); // Close file if we had one open
	myWad = Memory_free(myWad);
	return NULL;
}


// Replaces an existing file or adds it if new
fbool WadFile_Pack_File_As (struct wad_access_s* myWad, const char* diskFileName, const char* storeAsName)
{	
	if (myWad->fileMode == ARCHIVE_EDIT || myWad->fileMode == ARCHIVE_FASTEDIT )
	{
		// See if it is in file already
		int fileIndex = WadFile_Find_File_Index (myWad, storeAsName);
		
		if (fileIndex >=0)
			WadFile_Delete_File (myWad, storeAsName);
	}

	return WadFile_Add_Image_As (myWad, diskFileName, storeAsName);
}

fbool WadFile_Add_Image_As (struct wad_access_s* myWad, const char* diskFileName, const char* storeAsName)
{
	int i = myWad->totalEntries, stage, oldwritecursor = File_GetCursor (myWad->fileHandle);
	
	stage = CHECK_MODE_1;		if (myWad->fileMode == ARCHIVE_READ)				goto Wad_add_fail;
	stage = CHECK_LENGTH_2;		if (strlen(storeAsName) > MAX_WAD_FILENAME_IS_15)	goto Wad_add_fail;
	stage = CHECK_NUM_FILES_3;	if (myWad->totalEntries >= MAX_WAD_FILES_IS_4096)	goto Wad_add_fail;

//int			offset;				// Position of the entry in WAD
//int			wadsize;            // Size of the entry in WAD file
//int			memsize;            // Size of the entry in memory
//char		type;               // type of entry, should be 'D' (0x44) MipTex ('C' for Half-Life
//char		unused[3];          // Never compressed, so why bother
//char		name[16];           // 1 to 16 characters, '\0'-padded
	myWad->directory[i].filepos = File_GetCursor		  (myWad->fileHandle);  // LittleLong
	myWad->directory[i].memsize =
	myWad->directory[i].filelen = File_Append_BinaryFile (myWad->fileHandle, diskFileName); // This writes the file.   LittleLong


	myWad->directory[i].type	= myWad->isWAD3 ? 'C' : 'D'; // WAD3 = "C", WAD2 = "D";

	stage = CHECK_APPEND_4;		if (myWad->directory[i].filelen == 0)				goto Wad_add_fail;

	// If WAD3, appending padding + palette + padding
	if (myWad->isWAD3)
	{
		extern byte quakePalette [768];
		byte	pad[2] = {0, 1};

		myWad->directory[i].memsize += 2 + 768 + 2;
		myWad->directory[i].filelen += 2 + 768 + 2;
		File_Safe_Write (myWad->fileHandle, pad,		     sizeof(pad));			// 2
		File_Safe_Write (myWad->fileHandle, quakePalette, sizeof(quakePalette));	// 768
		File_Safe_Write (myWad->fileHandle, pad,		     sizeof(pad));			// 2		
	}

	
	// Everything OK
	StringLCopy (myWad->directory[i].name, storeAsName);	// Finalize the name
	myWad->isDirty	= True;									// Edit file needs written
	myWad->dataSegmentSize += myWad->directory[i].filelen;// Increase data segment count
	myWad->totalEntries ++;								// Total entries + 1

	return True;

Wad_add_fail:

	File_SetCursor	(myWad->fileHandle, oldwritecursor);	// Set the cursor back to where it was just in case
	printf ("Can't store '%s' as '%s': %s" , diskFileName, storeAsName, wad_add_error[stage]);
	return False;
}

fbool WadFile_Rename_File (struct wad_access_s* myWad, const char* storeAsName, const char* newName)
{
	int i = WadFile_Find_File_Index (myWad, storeAsName);
	
	if (i == -1)
	{
		printf ("Couldn't rename '%s' because couldn't find it in Wad\n", storeAsName);
		return False; // Couldn't find it so nothing to do.
	}

	if (myWad->fileMode != ARCHIVE_EDIT && myWad->fileMode != ARCHIVE_FASTEDIT)
	{
		printf ("Access mode is not EDIT and does not allow entry rename\n");
		return False; // Can't delete
	}
	
	if (strlen(newName) > MAX_WAD_FILENAME_IS_15)
	{
		printf ("New file name too long '%s' is %i characters (MAX %i)\n", newName, strlen(newName), MAX_WAD_FILENAME_IS_15);
		return False; // Rename 
	}

	StringLCopy (myWad->directory[i].name, newName);
	myWad->isDirty = True;
	return True;
}

fbool WadFile_Delete_File (struct wad_access_s* myWad, const char* storeAsName)
{
	int i = WadFile_Find_File_Index (myWad, storeAsName);
	
	if (i == -1) return False; // Nothing to do.

	if (myWad->fileMode != ARCHIVE_EDIT && myWad->fileMode != ARCHIVE_FASTEDIT)
	{
		printf ("Access mode is not EDIT and does not allow entry delete\n");
		return False; // Can't delete
	}
	
	myWad->directory[i].name[0] = '\0'; // Set name to zero length

	myWad->isDirty = True;
	return True;

}

// File cursor must be at end of file
struct wad_access_s* WadFile_Close (struct wad_access_s* myWad)
{
	fbool directoryWasCompressed	= WadFile_Close_Directory_Fixup (myWad); 
	fbool directoryWasAlphaSorted	= WadFile_Close_Directory_AlphaSort (myWad);
	fbool WadNeedsCompressed		= WadFile_Close_Check_For_Empty_Space (myWad);

	if (myWad->isDirty == True) // Store the offset then write the directory.
	{
		int directoryStart			= File_GetCursor (myWad->fileHandle);	
		int directoryByteLength		= sizeof(myWad->directory[0]) * myWad->totalEntries;

		myWad->header.id[0]		= 'W';		// Construct final header
		myWad->header.id[1]		= 'A';
		myWad->header.id[2]		= 'D';
		myWad->header.id[3]		= myWad->isWAD3 ? '3' : '2';
		myWad->header.numEntries	  = LittleLong(myWad->totalEntries);
		myWad->header.directoryOffset = LittleLong(directoryStart);
		
		
		File_Safe_Write			(myWad->fileHandle, &myWad->directory, directoryByteLength);  // Write directory
		File_SetCursor_ToStart	(myWad->fileHandle);											// Return to beginning of file
		File_Safe_Write			(myWad->fileHandle, &myWad->header, sizeof(myWad->header) ); // Write header
	}

	if (myWad->fileMode == ARCHIVE_EDIT && WadNeedsCompressed)  // Create can't need it, read can't do it, fastedit doesn't want it
		WadFile_Close_Compress (myWad);	// Recover deleted space. Re-writes whole file.

	fclose (myWad->fileHandle);  // close the file

	// Free the memory (this should effective return NULL, since Memory_free returns NULL)
	return Memory_free (myWad);
}

static const char* specialTextureExtensions[] =
{
	"_bump",
	"_gloss",
	"_glow",
	"_luma",
	"_norm",
};

static const int numSpecialTextureExtensions = sizeof(specialTextureExtensions) / sizeof(specialTextureExtensions[0]);




int WadFile_Misc_Is_Name_Special (const char* textureName)
{
	int i;
	for (i = 0; i < numSpecialTextureExtensions; i ++)					
		if (StringEndsWith (textureName, specialTextureExtensions[i]))
			return strlen(specialTextureExtensions[i]); // Special extension that does not go in a WAD

	return False;
}

//fbool WadFile_Misc_Valid_Special_Name (const char* textureName)
//{
//	int	specialNameLength = WadFile_Misc_Is_Name_Special (textureName);
//
//	if (specialNameLength && (strlen(textureName)-specialNameLength) <= MAX_WAD_FILENAME_IS_15 )
//		return True;
//
//	return False;
//}

fbool  WadFile_Misc_Is_Texture_Name_Too_Long (const char* textureName)
{
	int specialLength = WadFile_Misc_Is_Name_Special (textureName);
	int strlenTextureName = strlen (textureName);

// If the name isn't special and > 15
	if (specialLength == 0 && strlenTextureName > 15)
		return True; // Regular name too long

	if (specialLength > 0 && (strlenTextureName - specialLength) > 15)
		return True; // Special name too long

	return False; // Guess name is OK
}


void WadFile_Compress (const char* filename)
{
	// Just open and close in edit mode and viola!
	void* myPack = WadFile_Open (filename, ARCHIVE_EDIT);
	WadFile_Close (myPack);
}

//
// I'm not a 100% on the functions below here.
//
//
//fbool WadFile_Misc_Is_Name_Wadable (const char* textureName)
//{
//	int i;
//
//	if (strlen(textureName) > MAX_WAD_FILENAME_IS_15)
//		return False; // Name too long
//
//	
//	for (i = 0; i < numSpecialTextureExtensions; i ++)					
//		if (StringEndsWith (textureName, specialTextureExtensions[i]))
//			return False; // Special extension that does not go in a WAD
//
//	return True;
//}
//
//fbool WadFile_Misc_Is_Image_Size_Wadable (int width, int height)
//{
//	if (!width || !height)
//		return False;
//
//	if (width & 15 || height & 15)
//		return False; // Image must be divisible by 16
//
//	return True;
//}
//
//char* WadFile_Misc_Make_Wad_Name (char* outTextureName256, const char* sourceFileName)
//{
//	strlcpy						(outTextureName256, String_SkipPath(sourceFileName), 256);
//	String_Edit_StripExtension  (outTextureName256);
//
//	// NO ....
//	//if (strlen(sourceFileName) > MAX_WAD_FILENAME_IS_15)
//	//	printf ("This texture name '%s' cannot go into a WAD\n", outTextureName);
//	//	return outTextureName;
//
//	if (outTextureName256[0] = '#')	// Water texture
//		outTextureName256[0] = '*';
//
//	return outTextureName256;
//}
//
//fbool WadFile_Maybe_Add_Image_File (struct wad_access_s* myWad, const char* imageFileName)
//{
//	char		tempbuffer[256];
//	char*		myTempName	= WadFile_Misc_Make_Wad_Name   (tempbuffer, imageFileName);
//	fbool		isNameOK	= WadFile_Misc_Is_Name_Wadable (myTempName);
//	imageRGBA_t myImage;
//	fbool		wroteIt		= False;
//
//	if (!isNameOK)
//	{
//		printf ("Rejected: Name '%s' is not an acceptable WAD name\n", myTempName);	
//		return False;
//	}
//
//	Image_Import_FileX (&myImage,imageFileName);
//	do
//	{
//		// Reject for size?
//		if (WadFile_Misc_Is_Image_Size_Wadable (myImage.width, myImage.height) == False)
//		{
//			printf ("Image '%s' has unacceptable size ...\n", myTempName);
//			break;
//		}
//
//		// Reject for having an alpha channel?
//		if (Image_Info_AlphaType (&myImage) != COLORINFO_ALPHA_SOLID_WHITE)
//		{
//			printf ("Image '%s' has an alpha channel; skipping.  For now ...\n", myTempName);
//			break;
//		}
//
//		Image_Export_Mip_File (&myImage, "_miptex_temp.mip", myTempName);
//		WadFile_Add_Image_As (myWad, "_miptex_temp.mip", myTempName);
//		printf ("Wadded: '%s' ...\n", myTempName);
//		wroteIt = True;
//	} while (0);
//	Image_Destroy (&myImage);
//
//	return wroteIt;
//}
//

fbool WadFile_Is_Edit_Mode (struct wad_access_s* myWad)
{
	if (myWad->fileMode == ARCHIVE_EDIT || myWad->fileMode == ARCHIVE_FASTEDIT)
		return True;

	return False;
}