//**********************************************************************************************************\
//
//	image_resample.c:	Image resampling stuff
//
//**********************************************************************************************************/

#include "project.h"


float calc_and_Deplete_PaintShare (float* paintAmountRemaining, fbool isFirst, float firstFrac)
{
	float amountToPaint = 1;

	// Check for first fraction or last fraction exceptions 
	     if (*paintAmountRemaining < 1)	amountToPaint =  *paintAmountRemaining; // Last col/row uses all remaining
	else if (isFirst && firstFrac)		amountToPaint = firstFrac;				// First col/row is going to use its fraction

	// Deplete paintAmountRemaining
	*paintAmountRemaining = *paintAmountRemaining - amountToPaint;

	return amountToPaint;
}

// Mixes in red, green, blue and alpha components based on weight to paint and the weight of what has already been painted to the pixel.
void Canvas_PaintPixel (float* canvasFloat, int row, int col, int numCols, float paintAmount, float redPercent, float greenPercent, float bluePercent, float alphaPercent)
{
	float*	outBucket = canvasFloat + 5* (row * numCols + col);

	float	oldWeight = outBucket[4] / (paintAmount + outBucket[4]);	// Influence of previous contents
	float	newWeight = paintAmount  / (paintAmount + outBucket[4]);	// Influence of new contents being added

	// Ok paint ... Calculate weighted values and mix

	// Mix new values in
	outBucket[0] = outBucket[0] * oldWeight + redPercent   * newWeight;
	outBucket[1] = outBucket[1] * oldWeight + greenPercent * newWeight;
	outBucket[2] = outBucket[2] * oldWeight + bluePercent  * newWeight;
	outBucket[3] = outBucket[3] * oldWeight + alphaPercent * newWeight;
	
	outBucket[4] = outBucket[4] + paintAmount;
}


void Spray_Pixel_To_Canvas (unsigned pixelColor, int sourceRow, int sourceCol, float pixelWidth, float pixelHeight, float* floatCanvas, int canvasWidth, int canvasHeight)
{

	// Pixel area of influence
	float pixelStart_x	= pixelWidth * sourceCol;
	float pixelEnd_x	= pixelStart_x + pixelWidth;

	float pixelStart_y	= pixelHeight * sourceRow;
	float pixelEnd_y	= pixelStart_y + pixelHeight;

	// Pixel needs to paint a total amount of .... x on canvas, with y influence per row and z influence per col
	float totalGallonsOfPaint	= pixelWidth * pixelHeight;	// One pixel needs to paint this much on the canvas in total
	float virtualDestRow, virtualDestCol;

	// Color components
	float redPercent, greenPercent, bluePercent, alphaPercent;
	Pixel_RGBA_To_Components_Percents (pixelColor,  &redPercent, &greenPercent, &bluePercent, &alphaPercent);


	// ROW	
	{
		// Full bucket for row
		float gallonsPaintTopY		= pixelHeight;

		fbool isFirstRow			= True;
		float endRangeFirst			= floorf(pixelStart_y + 1);
		float firstRowFraction		= endRangeFirst - pixelStart_y;
		
		// Discrete mathematics: pixels are integers, convert float to appropriate discrete values.
		// If our range is 5.25 to 8.8 we will loop from 5 to 8.  Using .75, 1, 1, .8 of paint.
		for (virtualDestRow = floorf(pixelStart_y); virtualDestRow < pixelEnd_y; virtualDestRow ++)
		{
			float amountToUseRowPaint = calc_and_Deplete_PaintShare (&gallonsPaintTopY, isFirstRow, firstRowFraction);

			// Column
			{
				// Full bucket for row
				float gallonsPaintLeftX		= pixelWidth;

				fbool isFirstCol			= True;
				float firstColFraction		= floorf(pixelStart_x + 1) - pixelStart_x;			
				
				for (virtualDestCol = floorf(pixelStart_x); virtualDestCol < pixelEnd_x; virtualDestCol ++)
				{
					float amountToUseColPaint	= calc_and_Deplete_PaintShare (&gallonsPaintLeftX, isFirstCol, firstColFraction);

					// PIXEL
					{
						float combinedPaintUse	= amountToUseRowPaint * amountToUseColPaint;
						totalGallonsOfPaint		-= combinedPaintUse;

						Canvas_PaintPixel (floatCanvas, (int)virtualDestRow, (int)virtualDestCol, canvasWidth, combinedPaintUse, redPercent, greenPercent, bluePercent, alphaPercent);

					} // END PIXEL OUTER

					isFirstCol = False;

				} // END COLUMN LOOP

			} // END COLUMN OUTER

			isFirstRow = False;
		} // END ROW LOOP

	} // END ROW OUTER

	// FUNCTION DONE
}

// Actual resample



byte* Image_RGBA_Resample (const byte* source, const int sourceSize, const int sourceWidth, const int sourceHeight,
								 byte* dest,   const int destSize,   const int destWidth,   const int destHeight)
{
	int sourceRow, sourceCol, destRow, destCol;
	int destIndex = 0;
	
	unsigned* sourceImage		= (unsigned*) source;
	unsigned* destImage			= (unsigned*) dest;

	// Yes we are allocating 5 floats here: Red, green, blue, alpha and "share" to interpolate with high precision.
	float* floatCanvas			= Memory_calloc (1, destWidth * destHeight * sizeof(float) * 5, "Temp buffer to paint");

	// This is how much a source pixel affects a destination pixel in total
	const float pixelWeightX	= (float)destWidth  / sourceWidth;
	const float pixelWeightY	= (float)destHeight / sourceHeight;
	
	for (sourceRow = 0; sourceRow < sourceHeight; sourceRow ++)
		for (sourceCol = 0; sourceCol < sourceWidth; sourceCol ++)
		{
			unsigned	srcPixel	= sourceImage[sourceRow  * sourceWidth + sourceCol]; // Our pixel
			Spray_Pixel_To_Canvas (srcPixel, sourceRow, sourceCol, pixelWeightX, pixelWeightY, floatCanvas, destWidth, destHeight); 
		}

	// Convert Float Canvas to byte canvas
//
	for (destRow = 0; destRow < destHeight; destRow ++)
		for (destCol = 0; destCol < destWidth; destCol ++, destIndex++ )
//			destImage[destIndex] = Pixel_RGBA_From_Color4pct (	floatCanvas + destRow * destWidth * destCol );
		{
//			int destOffset = destIndex ;
			float*	outBucket = floatCanvas + 5 * (destRow * destWidth + destCol);
			destImage[destIndex] = Pixel_RGBA_From_Components_Percents (outBucket[0], outBucket[1], outBucket[2], outBucket[3]);

		}
// floatCanvas[destIndex] * 255; /
	// canvasFloat + row * numCols + col
	floatCanvas = Memory_free (floatCanvas); // Free and NULL
	return dest;
}

